@@ -16,7 +16,8 @@
 # Pyskool. If not, see <http://www.gnu.org/licenses/>.
 
 import random
-from skool import Location, Signals
+import signals
+from location import Location
 from animatorystates import *
 
 class CommandListTemplate:
@@ -38,12 +39,17 @@ class CommandList:
         self.template = None
         self.controlling_command = None
         self.controlling_command_timer = None
+        self.subcommand = None
 
     def command(self, character):
-        # If the controlling command is not already in control of the character,
-        # give it a chance
-        if self.controlling_command and self.controlling_command not in self.stack:
+        # If the controlling command is not already in control of the
+        # character, give it a chance (unless the current command is
+        # uninterruptible)
+        if self.stack and self.is_interruptible() and self.controlling_command and self.controlling_command not in self.stack:
             self.stack.append(self.controlling_command)
+        # Now give the subcommand a chance to run
+        if self.subcommand and self.subcommand not in self.stack and self.is_interruptible():
+            self.stack.append(self.subcommand)
         while True:
             if self.stack:
                 command = self.stack[-1]
@@ -52,7 +58,7 @@ class CommandList:
                     break
                 elif subcommand is not command:
                     self.stack.append(subcommand)
-                else:
+                elif self.stack:
                     self.stack.pop()
             else:
                 # Remove any controlling command before moving to the next
@@ -74,6 +80,7 @@ class CommandList:
 
     def set_template(self, template):
         self.template = template
+        self.subcommand = None
         while self.stack and self.stack[0].is_interruptible():
             self.stack.pop(0)
         self.restart_index = 0
@@ -109,6 +116,16 @@ class CommandList:
         if destination and self.is_GoToing():
             self.stack[0].set_destination(destination)
 
+    def is_interruptible(self):
+        for command in self.stack:
+            if not command.is_interruptible():
+                return False
+        return True
+
+    def set_subcommand(self, command_name, args):
+        command_class = eval(command_name)
+        self.subcommand = command_class(*args)
+
 class Command:
     def is_interruptible(self):
         return True
@@ -135,12 +152,12 @@ class SitForAssembly(ComplexCommand):
         return (self.find_spot_to_sit, self.sit_down, self.get_up, self.done)
 
     def find_spot_to_sit(self, character):
-        if character.got_signal(Signals.ASSEMBLY_FINISHED):
+        if character.got_signal(signals.ASSEMBLY_FINISHED):
             return self
         return GoToXY(character.x - random.randrange(1, 5), character.y)
 
     def sit_down(self, character):
-        if character.got_signal(Signals.ASSEMBLY_FINISHED):
+        if character.got_signal(signals.ASSEMBLY_FINISHED):
             return self
         if character.direction < 0:
             character.turn()
@@ -148,51 +165,50 @@ class SitForAssembly(ComplexCommand):
         character.sit()
 
     def get_up(self, character):
-        if character.got_signal(Signals.ASSEMBLY_FINISHED):
+        if character.got_signal(signals.ASSEMBLY_FINISHED):
             character.get_up()
         else:
             return False
 
 class MoveDoor(ComplexCommand):
-    def __init__(self, door_id, shut):
+    def __init__(self, barrier_id, shut):
         ComplexCommand.__init__(self)
-        self.door_id = door_id
+        self.barrier_id = barrier_id
         self.shut = shut
 
     def get_commands(self):
         return (self.raise_arm, self.move_door, self.lower_arm, self.done)
 
     def raise_arm(self, character):
-        if character.check_door_status(self.door_id, self.shut):
+        if character.check_door_status(self.barrier_id, self.shut):
             """Move on now if the door is already in the desired state."""
             return self
         character.raise_arm()
 
     def move_door(self, character):
-        character.move_door(self.door_id, self.shut)
+        character.move_door(self.barrier_id, self.shut)
 
     def lower_arm(self, character):
         character.lower_arm()
 
 class OpenDoor(MoveDoor):
-    def __init__(self, door_id):
-        MoveDoor.__init__(self, door_id, False)
+    def __init__(self, barrier_id):
+        MoveDoor.__init__(self, barrier_id, False)
 
 class ShutDoor(MoveDoor):
-    def __init__(self, door_id):
-        MoveDoor.__init__(self, door_id, True)
+    def __init__(self, barrier_id):
+        MoveDoor.__init__(self, barrier_id, True)
 
 class WipeBoard(ComplexCommand):
     def __init__(self):
         ComplexCommand.__init__(self)
         self.column = 7
-        self.interrupted = False
 
     def get_commands(self):
-        return (self.walk, self.wipe, self.lower_arm, self.walk_back, self.done)
+        return (self.walk, self.wipe, self.lower_arm, self.done)
 
     def walk(self, character):
-        character.set_wiping_board(True)
+        character.wiping_board = True
         return GoToXY(character.x - 1, character.y)
 
     def wipe(self, character):
@@ -205,16 +221,11 @@ class WipeBoard(ComplexCommand):
         if self.column >= 0:
             # Repeat until the board is clean
             self.restart()
-
-    def walk_back(self, character):
-        if self.interrupted:
-            # No need to walk to the middle of the board if we were interrupted
-            return self
-        character.set_wiping_board(False)
-        return GoToXY(character.x + 3, character.y)
+        else:
+            character.wiping_board = False
 
     def is_interruptible(self):
-        self.interrupted = True
+        # Teacher should finish the job even if the bell rings
         return False
 
 class GoTo(Command):
@@ -342,24 +353,39 @@ class GrassAndAnswerQuestions(Command):
         character.keep_seat()
         if not self.ready:
             self.ready = True
-            character.signal(Signals.SWOT_READY)
+            character.signal(signals.SWOT_READY)
             character.start_lesson()
         return character.next_swot_action()
 
-class StartAssemblyIfReady(Command):
+class WaitUntil(Command):
+    def __init__(self, signal):
+        self.signal = signal
+
     def command(self, character):
-        if character.got_signal(Signals.ASSEMBLY_FINISHED):
+        if character.got_signal(self.signal):
             return self
-        elif character.got_signal(Signals.TIME_FOR_ASSEMBLY):
+
+class DoAssemblyDuty(Command):
+    def __init__(self):
+        self.fetched_eric = False
+
+    def command(self, character):
+        if self.fetched_eric:
+            character.restart_table()
+            return self
+        if character.got_signal(signals.ASSEMBLY_FINISHED):
+            return self
+        if not character.is_time_to_start_lesson():
             return
-        character.restart_table()
-        return self
+        character.signal(signals.TIME_FOR_ASSEMBLY)
+        character.set_home_room()
+        if character.is_eric_absent():
+            self.fetched_eric = True
+            return FetchEric()
 
 class AwaitAssemblyTime(Command):
     def command(self, character):
-        if character.is_time_to_start_lesson():
-            character.signal(Signals.TIME_FOR_ASSEMBLY)
-        else:
+        if not character.got_signal(signals.TIME_FOR_ASSEMBLY):
             character.restart_table()
         return self
 
@@ -369,7 +395,8 @@ class ConductAssembly(Command):
 
     def command(self, character):
         if self.spoken:
-            character.signal(Signals.ASSEMBLY_FINISHED)
+            character.signal(signals.ASSEMBLY_FINISHED)
+            character.unset_home_room()
             return self
         self.spoken = True
         return Say(character.get_assembly_message())
@@ -384,7 +411,7 @@ class StartLessonIfReady(Command):
             if character.is_time_to_start_lesson():
                 self.ready = True
                 character.signal(self.signal)
-                character.reveal_secret(False)
+                character.reveal_safe_secret(False)
                 return TellKidsToSitDown()
             else:
                 character.restart_table()
@@ -409,20 +436,28 @@ class TellKidsToSitDown(Command):
         return Say(character.sit_down_message)
 
 class ConductClass(Command):
+    def command(self, character):
+        if character.is_teaching_eric():
+            if character.got_signal(signals.SWOT_READY):
+                return ConductClassWithEric()
+        else:
+            return ConductClassWithoutEric()
+
+class ConductClassWithoutEric(Command):
     def __init__(self):
         self.wiped_board = False
+        self.walked_to_board = False
         self.wrote_on_board = False
         self.told_class = False
 
     def command(self, character):
-        if character.is_teaching_eric() and not character.got_signal(Signals.SWOT_READY):
-            return
-        if character.is_teaching_eric():
-            return ConductClassWithEric()
         if character.room.has_blackboard():
             if not self.wiped_board:
                 self.wiped_board = True
                 return WipeBoard()
+            if not self.walked_to_board:
+                self.walked_to_board = True
+                return GoToXY(character.x + 3, character.y)
             if not self.wrote_on_board:
                 self.wrote_on_board = True
                 if random.randrange(256) >= 160:
@@ -541,7 +576,7 @@ class FireNowAndThen(HitOrFireNowAndThen):
             return False
         if character.x % 4 != 0:
             return False
-        return not character.pellet.is_airborne()
+        return not character.pellet.is_visible()
 
     def get_command(self):
         return FireCatapult()
@@ -552,13 +587,13 @@ class FireCatapult(ComplexCommand):
 
     def aim(self, character):
         if character.is_raising_catapult():
-            character.fire_catapult()
+            character.aim_catapult()
         else:
             character.raise_catapult()
             return False
 
     def fire(self, character):
-        character.fire()
+        character.fire_catapult()
 
     def lower(self, character):
         if character.is_lowering_catapult():
@@ -577,7 +612,7 @@ class MovePellet(Command):
         self.y_inc = 0
 
     def command(self, pellet):
-        if not pellet.is_airborne():
+        if not pellet.is_visible():
             return
         if self.count <= 0:
             self.count = pellet.pellet_range
@@ -585,14 +620,20 @@ class MovePellet(Command):
             self.y_inc = 0
         pellet.run(True)
         self.count -= 1
-        if self.count == 0 or pellet.barrier or pellet.check_shields():
-            return self.end_flight(pellet)
+        if self.count == 0 or pellet.barrier:
+            self.end_flight(pellet)
+            return
+        if self.y_inc < 0 or self.count <= pellet.hit_zone:
+            if pellet.hit_shield() or pellet.hit_cup() or pellet.hit_conker():
+                self.end_flight(pellet)
+                return
         if self.y_inc == 0 and self.count <= pellet.hit_zone:
             victim = pellet.get_victim()
             if victim:
                 if victim.is_deckable():
                     victim.deck()
-                    return self.end_flight(pellet)
+                    self.end_flight(pellet)
+                    return
                 if victim.is_adult() and victim.is_knocked_over():
                     self.x_inc = 0
                     self.y_inc = -1
@@ -607,12 +648,119 @@ class MovePellet(Command):
     def is_interruptible(self):
         return False
 
+class FireWaterPistol(ComplexCommand):
+    def get_commands(self):
+        return (self.aim, self.fire, self.lower, self.done)
+
+    def aim(self, character):
+        character.aim_water_pistol()
+
+    def fire(self, character):
+        character.fire_water_pistol()
+
+    def lower(self, character):
+        character.complete_action()
+
+    def is_interruptible(self):
+        return False
+
+class DumpWaterPistol(ComplexCommand):
+    def get_commands(self):
+        return (self.dump_water_pistol, self.stand_up)
+
+    def dump_water_pistol(self, eric):
+        eric.dump_water_pistol()
+
+    def stand_up(self, eric):
+        eric.stand_up()
+        return self
+
+class MoveWater(Command):
+    def __init__(self):
+        self.phase = 0
+        self.phases = ((WATER1, 2, -1, 0), (WATER2, 2, 0, 1), (WATER3, 1, 0, 0), (WATER4, 0, 1, 0), (WATER4, 0, 1, 2))
+
+    def command(self, water):
+        if not water.is_visible():
+            return
+        water.run(True)
+        state, x_inc, y_inc, hit = self.phases[self.phase]
+        if hit == 1 and water.hit_cup():
+            return self.end_flight(water)
+        if hit == 2 and water.hit_plant():
+            return self.end_flight(water)
+        if water.hit_floor():
+            return self.end_flight(water)
+        water.x += water.direction * x_inc
+        water.y += y_inc
+        water.animatory_state = state
+        self.phase = min(self.phase + 1, len(self.phases) - 1)
+
+    def end_flight(self, water):
+        self.phase = 0
+        water.hide()
+
+    def is_interruptible(self):
+        return False
+
+class DropStinkbomb(ComplexCommand):
+    def get_commands(self):
+        return (self.raise_arm, self.drop, self.lower, self.done)
+
+    def raise_arm(self, character):
+        character.raise_arm()
+
+    def drop(self, character):
+        character.drop_stinkbomb()
+
+    def lower(self, character):
+        character.complete_action()
+
+    def is_interruptible(self):
+        return False
+
+class Stink(Command):
+    def __init__(self):
+        self.reinitialise()
+
+    def reinitialise(self):
+        self.count = 29
+
+    def command(self, bomb):
+        if not bomb.is_visible():
+            return
+        self.count -= 1
+        if self.count <= 0:
+            self.reinitialise()
+            bomb.hide()
+            return
+        bomb.run(True)
+        bomb.move_cloud()
+
+class MoveDeskLid(ComplexCommand):
+    def __init__(self):
+        self.reinitialise()
+
+    def reinitialise(self):
+        self.count = 10
+
+    def command(self, desk_lid):
+        if not desk_lid.is_visible():
+            return
+        desk_lid.run(True)
+        if self.count >= 10:
+            desk_lid.deliver_contents()
+        self.count -= 1
+        if self.count <= 0:
+            self.reinitialise()
+            desk_lid.hide()
+
 class HitNowAndThen(HitOrFireNowAndThen):
     def ready(self, character):
         return HitOrFireNowAndThen.ready(self, character) and self.in_front_of_victim(character)
 
     def in_front_of_victim(self, character):
-        return len(character.get_facing_characters(3)) > 0
+        return len(character.get_punchees(3)) > 0
 
     def get_command(self):
         return Hit()
@@ -629,7 +777,7 @@ class Hit(ComplexCommand):
             return False
 
     def hit(self, character):
-        facing_characters = character.get_facing_characters(2)
+        facing_characters = character.get_punchees(2)
         if facing_characters:
             facing_characters[0].deck()
 
@@ -666,14 +814,19 @@ class Dethroned(Floored):
         Floored.__init__(self, 8)
 
 class KnockedOver(Floored):
-    def __init__(self):
+    def __init__(self, sleep):
         Floored.__init__(self, 15)
+        self.sleep = sleep
 
     def command(self, character):
+        if self.sleep:
+            if character.is_time_to_wake():
+                return self
+            return
         if Floored.command(self, character) is self:
             return self
         if self.count == 14:
-            character.reveal_secret(True)
+            character.reveal_safe_secret(True)
         if self.count == 7:
             character.reprimand()
 
@@ -716,6 +869,9 @@ class JumpIfOpen(Command):
 class WalkFast(Command):
     def command(self, character):
         character.run(True)
+        # Ensure that no speed change is triggered while this command is in
+        # effect
+        character.speed_change_delay = 10
         return self
 
 class StalkAndHit(HitNowAndThen):
@@ -755,13 +911,18 @@ class Write:
         self.finished = character.write()
 
 class FindEric(Command):
+    def __init__(self):
+        self.running = False
+
     def command(self, character):
         character.stop_clock()
         if character.is_beside_eric():
             if character.freeze_eric():
                 return self
             return
-        character.run(True)
+        if not self.running:
+            self.running = True
+            character.set_controlling_command(WalkFast())
         return GoTowardsXY(*character.get_location_of_eric())
 
 class Freeze:
@@ -781,29 +942,24 @@ class TellEric(Command):
         return Say(character.expand_names(self.message))
 
 class TellEricAndWait(Command):
+    DELAY = 20
+
     def __init__(self, message):
         self.told_eric = 0
-        self.delay = 20
+        self.delay = TellEricAndWait.DELAY
         self.message = message
 
     def command(self, character):
         if self.told_eric == 0:
             self.told_eric = 1
             return Say(character.expand_names(self.message))
-        elif self.told_eric == 1:
-            if character.eric_understood():
-                character.unfreeze_eric()
-                return self
-            self.told_eric = 2
-            return Say(character.get_understand_message())
-        elif self.told_eric == 2:
-            if character.eric_understood():
-                character.unfreeze_eric()
-                return self
-            self.delay -= 1
-            if self.delay < 0:
-                self.delay = 20
-                self.told_eric = 0
+        if character.eric_understood():
+            character.unfreeze_eric()
+            return self
+        self.delay -= 1
+        if self.delay < 0:
+            self.delay = TellEricAndWait.DELAY
+            self.told_eric = 0
 
 class SetClock(Command):
     def __init__(self, ticks):
@@ -839,13 +995,20 @@ class AddLines(Command):
         return False
 
 class FetchEric(Command):
+    def __init__(self):
+        self.running = False
+
     def command(self, character):
         if not character.is_eric_absent() or character.is_eric_expelled():
             character.run(False)
             return self
-        if not (character.is_beside_eric() and character.is_facing_eric()):
-            character.run(True)
+        if not self.running:
+            self.running = True
+            character.set_controlling_command(WalkFast())
+        if not character.is_beside_eric():
             return GoTowardsXY(*character.get_location_of_eric())
+        if not character.is_facing_eric():
+            character.turn()
 
 class FindEricIfMissing(Command):
     def command(self, character):
@@ -856,6 +1019,7 @@ class FindEricIfMissing(Command):
 class TripPeopleUp(Command):
     def command(self, character):
         character.run(True)
+        character.speed_change_delay = 10
         character.trip_people_up()
         return self
 
@@ -873,8 +1037,9 @@ class Follow(Command):
 class MoveMouse(Command):
     def __init__(self):
         self.reset_sprints()
-        self.hide_x = None
+        self.hide_x = -1
         self.hide_delay = None
+        self.life = random.randint(10, 41)
 
     def reset_sprints(self):
         self.sprints = random.randint(2, 5)
@@ -885,36 +1050,89 @@ class MoveMouse(Command):
 
     def command(self, mouse):
         mouse.run(True)
-        if self.hide_x:
+        if self.hide_x >= 0:
+            if not mouse.immortal:
+                self.life -= 1
+                if self.life <= 0:
+                    mouse.die()
+                    return
             if self.hide_delay > 0:
                 self.hide_delay -= 1
                 return
             mouse.x = self.hide_x
-            self.hide_x = None
+            self.hide_x = -1
             self.reset_sprints()
         elif self.sprint_distance > 0:
-            if mouse.is_blocked():
-                mouse.turn()
-            mouse.x += mouse.direction
             self.sprint_distance -= 1
+            self.move(mouse)
         elif self.sprints > 0:
             self.sprints -= 1
             self.reset_sprint_distance()
             mouse.direction = random.choice((-1, 1))
-            if mouse.is_blocked():
-                mouse.turn()
-            mouse.x += mouse.direction
+            self.move(mouse)
         else:
             self.hide_x = mouse.x
             self.hide_delay = random.randint(5, 20)
             mouse.hide()
 
+    def move(self, mouse):
+        if mouse.is_blocked():
+            mouse.turn()
+        mouse.x += mouse.direction
+        mouse.scare_people()
+
+    def is_interruptible(self):
+        return False
+
+class EvadeMouse(Command):
+    def __init__(self):
+        self.count = 21
+        self.in_the_air = False
+        self.on_a_chair = False
+        self.old_as = None
+
+    def command(self, character):
+        if self.old_as is None:
+            self.old_as = character.animatory_state
+            if character.is_sitting():
+                character.get_up()
+                return
+        self.count -= 1
+        if self.count <= 0:
+            if self.in_the_air or self.on_a_chair:
+                character.y += 1
+                self.in_the_air = self.on_a_chair = False
+                return
+            character.animatory_state = self.old_as
+            return self
+        character.run(False)
+        if self.in_the_air:
+            character.y += 1
+            self.in_the_air = False
+        elif not self.on_a_chair:
+            if character.chair(None):
+                self.on_a_chair = True
+            character.y -= 1
+            self.in_the_air = not self.on_a_chair
+
     def is_interruptible(self):
         return False
 
 class MoveFrog(Command):
     def command(self, frog):
-        if frog.x < 0:
+        if frog.trapped or not frog.is_visible():
+            return
+        if frog.falling:
+            if frog.check_heads():
+                # Make the frog bounce off the character's head
+                frog.animatory_state = HOP1
+                frog.direction = -1
+                frog.x -= 1
+            else:
+                frog.y += 1
+                if frog.get_floor():
+                    frog.falling = False
+                    frog.animatory_state = SIT
             return
         if not frog.is_eric_nearby() and random.randint(0, 3):
             return
@@ -953,20 +1171,315 @@ class Hop(Command):
 
 class Catch(ComplexCommand):
     def get_commands(self):
-        return (self.bend_over, self.catch_animal, self.stand_up, self.done)
+        return (self.catch_animal, self.stand_up)
 
-    def bend_over(self, character):
-        character.bend_over()
+    def catch_animal(self, eric):
+        eric.catch_animal()
 
-    def catch_animal(self, character):
-        character.catch_animal()
+    def stand_up(self, eric):
+        eric.stand_up()
+        return self
 
-    def stand_up(self, character):
-        character.stand_up()
+class ReleaseMice(ComplexCommand):
+    def get_commands(self):
+        return (self.release_mice, self.stand_up)
+
+    def release_mice(self, eric):
+        eric.release_mice()
+
+    def stand_up(self, eric):
+        eric.stand_up()
+        return self
+
+class Grow(Command):
+    BEGIN = 26
+    APPEAR = 14
+    END = 7
+
+    def __init__(self):
+        self.count = Grow.BEGIN
+
+    def command(self, plant):
+        if not plant.growing:
+            return
+        plant.run(True)
+        self.count -= 1
+        if self.count == Grow.APPEAR:
+            plant.appear()
+        elif self.count == Grow.END:
+            plant.finish_growing()
+        elif self.count <= 0:
+            plant.die()
+            self.count = Grow.BEGIN
+
+    def is_interruptible(self):
+        return False
+
+class Fall(Command):
+    def command(self, thing):
+        if not thing.is_visible():
+            return
+        thing.run(True)
+        if thing.floor or thing.hit_victim():
+            thing.hide()
+        thing.y += 1
+
+    def is_interruptible(self):
+        return False
+
+class MoveBike(Command):
+    def command(self, bike):
+        if bike.momentum <= 0 or not bike.is_visible():
+            return
+        bike.wheel()
+        if bike.momentum <= 0:
+            bike.fall()
+
+class RideBike(Command):
+    def __init__(self, bike):
+        self.bike = bike
+        bike.prepare()
 
-#
-# Unimplemented commands
-#
-class WatchForEric(Command):
     def command(self, character):
-        return
+        character.check_bike_keys()
+        if character.rode_into_barrier(self.bike):
+            self.bike.momentum = 0
+        elif character.pedalled():
+            self.bike.pedal()
+            return
+        elif character.dismounted():
+            character.stand_up()
+            self.bike.start_wheeling(character)
+            character.dismount()
+            return self
+        elif character.sitting_on_saddle:
+            if character.stood_on_saddle():
+                self.bike.start_wheeling(character)
+                character.stand_on_saddle()
+        elif character.got_back_on_saddle():
+            character.get_back_on_saddle()
+            self.bike.hide()
+        elif character.jumped():
+            character.dismount()
+            return JumpOffSaddle()
+        else:
+            # Eric is standing on the saddle of the bike; move him with it
+            character.x += character.direction
+        self.bike.momentum -= 1
+        if self.bike.momentum <= 0:
+            character.dismount()
+            if character.sitting_on_saddle:
+                self.bike.fall(character)
+            else:
+                self.bike.fall()
+                if character.rode_into_barrier(self.bike, 'SkoolGate'):
+                    return FlyOverSkoolGate()
+            return FallToFloor()
+
+class FallToFloor(Command):
+    def __init__(self, land_on_feet=False):
+        self.land_on_feet = land_on_feet
+
+    def command(self, character):
+        if character.floor:
+            if not self.land_on_feet:
+                character.fall_to_floor()
+            return self
+        character.y += 1
+
+class Flight(Command):
+    def __init__(self, phases, paralyse=False):
+        self.index = 0
+        self.phases = phases
+        self.paralyse = paralyse
+
+    def command(self, eric):
+        if self.index < len(self.phases):
+            x_inc, y_inc, state = self.phases[self.index]
+            self.index += 1
+            eric.fly(x_inc, y_inc)
+            eric.animatory_state = state
+            return
+        eric.end_flight()
+        if self.paralyse:
+            eric.paralyse()
+            return Freeze()
+        return self
+
+class FlyOverSkoolGate(Flight):
+    def __init__(self):
+        phases = []
+        phases.append((1, -1, WALK0))
+        phases.append((1, -1, WALK0))
+        phases.append((1, -1, WALK0))
+        phases.append((1, 1, WALK0))
+        phases.append((1, 1, SITTING_ON_CHAIR))
+        phases.append((1, 0, SITTING_ON_FLOOR))
+        phases.append((1, 1, SITTING_ON_FLOOR))
+        phases.append((1, 1, SITTING_ON_FLOOR))
+        Flight.__init__(self, phases)
+
+class JumpOutOfWindow(Flight):
+    def __init__(self, window_id):
+        paralyse = False
+        phases = []
+        phases.append((0, 0, WALK1))
+        phases.append((1, 0, WALK2))
+        phases.append((0, 0, WALK3))
+        phases.append((1, 0, WALK0))
+        if window_id == 'MiddleWindow':
+            phases.extend([(0, 1, WALK0)] * 10)
+        elif window_id == 'UpperWindow':
+            paralyse = True
+            phases.extend([(1, 1, WALK0)] * 4)
+            phases.extend([(0, 1, WALK0)] * 6)
+            phases.extend([(0, 1, SITTING_ON_CHAIR)] * 3)
+            phases.extend([(0, 1, SITTING_ON_FLOOR)] * 3)
+            phases.append((0, 1, KNOCKED_OUT))
+        Flight.__init__(self, phases, paralyse)
+
+class ClimbOverSkoolGate(Flight):
+    def __init__(self):
+        phases = []
+        phases.append((0, 0, WALK1))
+        phases.append((1, 0, WALK2))
+        phases.append((0, 0, WALK3))
+        phases.append((1, 0, WALK0))
+        phases.append((0, 0, WALK1))
+        phases.append((1, 0, WALK2))
+        phases.extend([(0, 1, WALK2)] * 3)
+        Flight.__init__(self, phases)
+
+class JumpOffSaddle(ComplexCommand):
+    def get_commands(self):
+        return (self.rise, self.reach, self.check_cup, self.fall, self.done)
+
+    def rise(self, eric):
+        eric.y -= 1
+
+    def reach(self, eric):
+        eric.y -= 1
+        eric.animatory_state = ARM_UP
+
+    def check_cup(self, eric):
+        eric.check_cup()
+        eric.animatory_state = SITTING_ON_FLOOR
+
+    def fall(self, eric):
+        if eric.floor:
+            eric.fall_to_floor()
+            return
+        eric.y += 1
+        return False
+
+class Kiss(ComplexCommand):
+    def __init__(self):
+        ComplexCommand.__init__(self)
+        self.kissee = None
+
+    def get_commands(self):
+        return (self.start_kiss, self.finish_kiss, self.done)
+
+    def start_kiss(self, eric):
+        self.kissee = eric.kissee()
+        if self.kissee:
+            self.kissee.pause()
+            if self.kissee.will_kiss_eric():
+                eric.x += eric.direction
+                eric.hide()
+                self.kissee.kiss_eric()
+            else:
+                self.kissee.raise_arm()
+        else:
+            eric.walk()
+            eric.walk_delay = 4
+
+    def finish_kiss(self, eric):
+        if self.kissee:
+            if self.kissee.is_kissing_eric():
+                eric.kiss()
+                eric.unhide()
+                self.kissee.finish_kiss()
+            else:
+                self.kissee.lower_arm()
+                eric.deck()
+            self.kissee.resume()
+        else:
+            eric.stand_up()
+
+class Pause(Command):
+    def command(self, character):
+        if not character.paused:
+            return self
+
+class WatchForEric(Command):
+    def __init__(self, command_list_id, alert_message):
+        self.command_list_id = command_list_id
+        self.alert_message = alert_message
+
+    def command(self, character):
+        if character.should_stop_eric():
+            character.raise_arm()
+            character.stop_eric = True
+            return StopEric(self.command_list_id, self.alert_message)
+        return self
+
+class StopEric(Command):
+    def __init__(self, command_list_id, alert_message):
+        self.raised_alarm = False
+        self.command_list_id = command_list_id
+        self.alert_message = alert_message
+
+    def command(self, character):
+        if not self.raised_alarm:
+            self.raised_alarm = True
+            character.raise_alarm(self.alert_message, self.command_list_id)
+        if not character.should_stop_eric():
+            character.lower_arm()
+            character.stop_eric = False
+            return self
+
+class ShadowEric(Command):
+    def __init__(self):
+        self.running = False
+
+    def command(self, character):
+        if not self.running:
+            self.running = True
+            character.set_controlling_command(WalkFast())
+        if not character.is_beside_eric():
+            return GoTowardsXY(*character.get_location_of_eric())
+        if not character.is_facing_eric():
+            character.turn()
+
+class SetSubcommand(Command):
+    def __init__(self, command_name, *args):
+        self.command_name = command_name
+        self.args = args
+
+    def command(self, character):
+        character.set_subcommand(self.command_name, self.args)
+        return self
+
+class MonitorEric(Command):
+    def __init__(self, command_list_id):
+        self.command_list_id = command_list_id
+
+    def command(self, character):
+        if character.should_chase_eric():
+            character.change_command_list(self.command_list_id)
+        return self
+
+class ChaseEricOut(Command):
+    def __init__(self):
+        self.running = False
+
+    def command(self, character):
+        if not self.running:
+            self.running = True
+            character.set_controlling_command(WalkFast())
+        if character.x >= 137:
+            if not character.is_beside_eric():
+                return GoTowardsXY(*character.get_location_of_eric())
+            if not character.is_facing_eric():
+                character.turn()
