@@ -25,6 +25,10 @@ import keys
 
 LINES_DELAY = 60
 
+# The command list Mr Wacker should use to expel Eric after he's jumped out of
+# the top-floor window
+NOT_A_BIRD = 'ExpelEric:NotABird'
+
 class Eric(Character):
     def __init__(self, character_id, name, flags):
         Character.__init__(self, character_id, name, flags)
@@ -40,6 +44,7 @@ class Eric(Character):
         self.understood = False
         self.controller = None
         self.lines_delay = LINES_DELAY
+        self.last_lines_giver = None
         self.inventory = set()
         self.mice = 0
         self.last_bike_key = None
@@ -47,6 +52,8 @@ class Eric(Character):
         self.on_bike = False
         self.started_pedalling = False
         self.sitting_on_saddle = False
+        self.expelled = False
+        self.hide_coords = None
 
     def move(self, keyboard):
         self.keyboard = keyboard
@@ -54,7 +61,7 @@ class Eric(Character):
         self.walk_delay -= 1
         if self.walk_delay > 0:
             return 0
-        self.barrier = self.skool.barrier(self)
+        self.barrier = self.skool.barrier(self) or self.cast.get_eric_stopper()
         self.floor = self.get_floor()
         self.keyboard.pump()
         self.walk_delay = 2
@@ -69,56 +76,55 @@ class Eric(Character):
             self.make_walking_sound(self.get_walk_state_index())
             return self.walk()
         if self.is_knocked_out() or self.is_sitting():
-            if self.keyboard.was_pressed(keys.SIT_STAND):
+            if self.keyboard.was_pressed(*keys.SIT_STAND):
                 if self.is_sitting_on_chair():
                     self.chair().vacate()
                 self.animatory_state = WALK0
                 self.beeper.make_sitting_sound()
-            elif self.is_sitting_on_chair() and self.keyboard.was_pressed(keys.OPEN_DESK):
+            elif self.is_sitting_on_chair() and self.keyboard.was_pressed(*keys.OPEN_DESK):
                 self.open_desk(self.skool.desk(self))
             return 0
-        if self.keyboard.was_pressed(keys.SIT_STAND):
-            self.sit()
+        if self.keyboard.was_pressed(*keys.SIT_STAND) and self.sit():
             self.beeper.make_sitting_sound()
             return 0
-        if self.keyboard.was_pressed(keys.FIRE_CATAPULT) and self.can_fire_catapult():
+        if self.keyboard.was_pressed(*keys.FIRE_CATAPULT) and self.can_fire_catapult():
             self.controller = FireCatapult()
             return 0
-        if self.keyboard.was_pressed(keys.FIRE_WATER_PISTOL) and self.can_fire_water_pistol():
+        if self.keyboard.was_pressed(*keys.FIRE_WATER_PISTOL) and self.can_fire_water_pistol():
             self.controller = FireWaterPistol()
             return 0
-        if self.keyboard.was_pressed(keys.DUMP_WATER_PISTOL) and self.can_dump_water_pistol():
+        if self.keyboard.was_pressed(*keys.DUMP_WATER_PISTOL) and self.can_dump_water_pistol():
             self.bend_over()
             self.controller = DumpWaterPistol()
             return 0
-        if self.keyboard.was_pressed(keys.DROP_STINKBOMB) and self.can_drop_stinkbomb():
+        if self.keyboard.was_pressed(*keys.DROP_STINKBOMB) and self.can_drop_stinkbomb():
             self.controller = DropStinkbomb()
             return 0
-        if self.keyboard.was_pressed(keys.HIT):
+        if self.keyboard.was_pressed(*keys.HIT):
             self.controller = Hit()
             return 0
-        if self.keyboard.was_pressed(keys.JUMP):
+        if self.keyboard.was_pressed(*keys.JUMP):
             self.controller = Jump()
             return 0
-        if self.keyboard.was_pressed(keys.WRITE) and self.skool.beside_blackboard(self):
+        if self.keyboard.was_pressed(*keys.WRITE) and self.skool.beside_blackboard(self):
             self.raise_arm()
             self.controller = Write()
             self.keyboard.start_writing()
             self.writing = True
             return 0
-        if self.keyboard.was_pressed(keys.CATCH) and self.can_bend_over():
+        if self.keyboard.was_pressed(*keys.CATCH) and self.can_bend_over():
             self.bend_over()
             self.controller = Catch()
             return 0
-        if self.keyboard.was_pressed(keys.MOUNT_BIKE) and self.can_mount_bike():
+        if self.keyboard.was_pressed(*keys.MOUNT_BIKE) and self.can_mount_bike():
             self.mount_bike()
             self.controller = RideBike(self.cast.bike)
             return 0
-        if self.keyboard.was_pressed(keys.KISS) and self.can_kiss():
+        if self.keyboard.was_pressed(*keys.KISS) and self.can_kiss():
             self.previous_as = self.animatory_state
             self.controller = Kiss()
             return 0
-        if self.keyboard.was_pressed(keys.RELEASE_MICE) and self.can_release_mice():
+        if self.keyboard.was_pressed(*keys.RELEASE_MICE) and self.can_release_mice():
             self.bend_over()
             self.controller = ReleaseMice()
             return 0
@@ -128,14 +134,14 @@ class Eric(Character):
             return 0
         old_ws, old_as, old_direction = self.get_walk_state_index(), self.animatory_state, self.direction
         plant = self.cast.plant(self)
-        if self.keyboard.pressed(keys.LEFT):
+        if self.keyboard.pressed(*keys.LEFT):
             if plant and plant.is_fully_grown():
                 barrier = self.skool.barrier(self, 1)
                 if barrier and barrier.barrier_id == 'SkoolGate' and barrier.is_shut():
                     self.controller = ClimbOverSkoolGate()
                     return 0
             self.left()
-        elif self.keyboard.pressed(keys.RIGHT):
+        elif self.keyboard.pressed(*keys.RIGHT):
             if plant and plant.is_fully_grown():
                 window = self.skool.window(self)
                 if window and not window.is_shut():
@@ -145,9 +151,9 @@ class Eric(Character):
                     self.controller = ClimbOverSkoolGate()
                     return 0
             self.right()
-        elif self.keyboard.pressed(keys.UP):
+        elif self.keyboard.pressed(*keys.UP):
             self.up()
-        elif self.keyboard.pressed(keys.DOWN):
+        elif self.keyboard.pressed(*keys.DOWN):
             self.down()
         if (self.animatory_state, self.direction) != (old_as, old_direction):
             self.make_walking_sound(old_ws)
@@ -179,24 +185,49 @@ class Eric(Character):
 
     def get_lines(self):
         self.lines_delay = max(0, self.lines_delay - 1)
-        if self.lines_delay > 0:
+        if self.lines_delay > LINES_DELAY / 2:
             return
-        gave_lines = False
+        if self.lines_delay == 0:
+            self.last_lines_giver = None
+        room = self.skool.room(self)
         home_room = self.skool.get_home_room()
-        if self.is_sitting_on_stairs():
-            gave_lines = self.alert_lines_givers(lines.NO_SITTING_ON_STAIRS)
-        elif self.skool.in_no_go_zone(self):
-            gave_lines = self.alert_lines_givers(lines.GET_OUT)
+        if self.skool.in_no_go_zone(self):
+            self.alert_lines_givers(lines.GET_OUT, True)
         elif self.should_get_along():
-            gave_lines = self.alert_lines_givers(lines.GET_ALONG, True)
-        elif home_room and home_room.contains(self) and home_room.chairs and not self.is_sitting_on_chair():
-            gave_lines = self.alert_lines_givers(lines.SIT_DOWN)
-        elif self.is_sitting_on_floor() or self.is_knocked_out():
-            gave_lines = self.alert_lines_givers(lines.GET_UP)
-        elif self.on_bike:
-            gave_lines = self.alert_lines_givers(lines.NO_BIKES)
-        if gave_lines:
-            self.lines_delay = LINES_DELAY
+            self.alert_lines_givers(lines.GET_ALONG, True)
+        elif self.is_sitting_on_stairs():
+            self.alert_lines_givers(lines.NO_SITTING_ON_STAIRS, True)
+        elif self.is_knocked_out():
+            self.alert_lines_givers(lines.GET_UP, True)
+        elif self.on_bike and not self.skool.in_playground(self):
+            self.alert_lines_givers(lines.NO_BIKES, True)
+        elif self.skool.plant_pot(self.x, self.y) or self.cast.plant(self):
+            self.alert_lines_givers(lines.GET_OFF_PLANT, True)
+        else:
+            is_assembly = self.skool.is_assembly()
+            assembly_hall_id = 'AssemblyHall'
+            if self.is_sitting_on_floor():
+                # Eric is sitting on the floor...
+                if room and room.room_id == assembly_hall_id:
+                    if self.direction < 0:
+                        # ...facing left in the assembly hall...
+                        if is_assembly:
+                            # ...and it's assembly - tell him to sit facing the
+                            # stage
+                            self.alert_lines_givers(lines.SIT_FACING_STAGE, True)
+                        else:
+                            # ...and it's not assembly - tell him to get off
+                            # the floor
+                            self.alert_lines_givers(lines.GET_UP, True)
+                else:
+                    # ...somewhere other than the assembly hall - tell him to
+                    # get off the floor
+                    self.alert_lines_givers(lines.GET_UP, True)
+            elif self.is_standing() and home_room and home_room.contains(self) and (home_room.chairs or home_room.room_id == assembly_hall_id):
+                # Eric is standing up either in a classroom after the lesson
+                # has started, or in the assembly hall after assembly has
+                # started - tell him to sit down
+                self.alert_lines_givers(lines.SIT_DOWN, True)
 
     def is_absent(self):
         """Return whether Eric is playing truant."""
@@ -229,15 +260,23 @@ class Eric(Character):
             self.controller = None
         self.beeper.make_knocked_out_sound()
 
-    def alert_lines_givers(self, message_id, truant=False):
+    def alert_lines_givers(self, message_id, reset_delay=False):
         nearby_lines_givers = self.cast.get_nearby_lines_givers(self)
         if nearby_lines_givers:
-            teacher = self.skool.get_teacher()
-            if truant and teacher in nearby_lines_givers:
-                teacher.give_lines(self.character_id, teacher.get_come_along_message_id())
-            else:
-                nearby_lines_givers[0].give_lines(self.character_id, message_id)
-        return len(nearby_lines_givers) > 0
+            lines_giver = nearby_lines_givers[0]
+            if message_id == lines.GET_ALONG:
+                teacher = self.skool.get_teacher()
+                if teacher in nearby_lines_givers:
+                    lines_giver = teacher
+                    message_id = teacher.get_come_along_message_id()
+                elif self.skool.in_playground(self):
+                    message_id = lines.BACK_TO_SKOOL
+            if reset_delay:
+                if lines_giver is self.last_lines_giver:
+                    return
+                self.lines_delay = LINES_DELAY
+            lines_giver.give_lines(self.character_id, message_id)
+            self.last_lines_giver = lines_giver
 
     def aim_catapult(self):
         Character.aim_catapult(self)
@@ -296,6 +335,10 @@ class Eric(Character):
 
     def collect_desk_contents(self, desk):
         if desk.contents:
+            if desk.contents == items.STINKBOMBS3:
+                for item in (items.STINKBOMBS2, items.STINKBOMBS1):
+                    if item in self.inventory:
+                        self.inventory.remove(item)
             self.inventory.add(desk.contents)
             if desk.contents == items.WATER_PISTOL:
                 desk.empty()
@@ -325,6 +368,7 @@ class Eric(Character):
         animal = self.cast.get_animal(self)
         if animal:
             if animal.is_mouse():
+                self.skool.add_to_score(10)
                 self.beeper.make_mouse_sound()
                 self.mice += 1
                 self.print_mouse_inventory()
@@ -358,12 +402,19 @@ class Eric(Character):
 
     def get_location(self):
         """Return the on-floor location that is closest to Eric."""
+        if self.hide_coords:
+            # If Eric is hiding (e.g. while kissing), temporarily restore his
+            # pre-hide coordinates for the location check
+            self.x, self.y = self.hide_coords
         staircase = self.skool.staircase(self)
         if staircase and staircase.supports(self):
             if staircase.bottom.y - self.y < self.y - staircase.top.y:
                 return staircase.bottom.coords()
             return staircase.top.coords()
-        return self.x, self.skool.floor_below(self).y
+        location = (self.x, self.skool.floor_below(self).y)
+        if self.hide_coords:
+            self.hide()
+        return location
 
     def get_location_above_hand(self):
         return (self.x + self.direction + 1, self.y)
@@ -397,7 +448,7 @@ class Eric(Character):
         key_down_events = self.keyboard.key_down_events
         if key_down_events:
             first_event = key_down_events[0]
-            if first_event.key == keys.ENTER:
+            if first_event.key in keys.ENTER:
                 self.raise_arm()
                 self.keyboard.finish_writing()
                 return
@@ -410,10 +461,11 @@ class Eric(Character):
 
     def check_combinations(self, blackboard):
         if self.skool.got_bike_combination(blackboard):
+            self.skool.add_to_score(100)
             self.skool.unchain_bike()
             self.beeper.make_bingo_sound()
         elif self.skool.got_storeroom_combination(blackboard):
-            self.cast.release_frogs()
+            self.skool.add_to_score(100)
             self.inventory.add(items.STOREROOM_KEY)
             self.print_inventory()
             self.beeper.make_bingo_sound()
@@ -434,13 +486,17 @@ class Eric(Character):
             self.understood = False
 
     def check_understanding(self):
-        if self.keyboard.was_pressed(keys.UNDERSTOOD):
+        if self.keyboard.was_pressed(*keys.UNDERSTOOD):
             self.understood = True
 
     def understood_message(self):
         return self.understood
 
+    def paralyse(self):
+        self.cast.expel_eric(NOT_A_BIRD)
+
     def take_safe_key(self):
+        self.skool.add_to_score(200)
         self.beeper.make_safe_key_sound()
         self.inventory.add(items.SAFE_KEY)
         self.print_inventory()
@@ -455,6 +511,15 @@ class Eric(Character):
         self.skool.add_lines(-100)
         self.beeper.make_bingo_sound()
 
+    def hide(self):
+        self.hide_coords = (self.x, self.y)
+        Character.hide(self)
+
+    def unhide(self):
+        if self.hide_coords:
+            self.x, self.y = self.hide_coords
+            self.hide_coords = None
+
     def can_mount_bike(self):
         return RIDING_BIKE0 in self.as_dict_L and self.cast.is_beside_bike(self)
 
@@ -469,21 +534,25 @@ class Eric(Character):
     def check_bike_keys(self):
         self.walk_delay = bike.SPEED
         self.bike_key = None
-        for key in (keys.LEFT, keys.RIGHT, keys.UP, keys.DOWN, keys.JUMP):
+        for key in keys.LEFT + keys.RIGHT + keys.UP + keys.DOWN + keys.JUMP:
             if self.keyboard.was_pressed(key):
                 self.bike_key = key
                 break
 
-    def rode_into_barrier(self, barrier_id=None):
-        if self.barrier:
-            if barrier_id:
-                return self.barrier.barrier_id == barrier_id
-            return True
-        return False
+    def rode_into_barrier(self, bike, barrier_id=None):
+        if bike.is_visible():
+            return bike.hit_barrier(barrier_id)
+        # Check from a distance of 1 if travelling leftwards so that the bike
+        # does not appear too close to the barrier if it hits it
+        distance = 0 if self.direction > 0 else 1
+        barrier = self.skool.barrier(self, distance)
+        if barrier_id:
+            return barrier and barrier.barrier_id == barrier_id
+        return barrier is not None
 
     def pedalled(self):
         if self.sitting_on_saddle:
-            pedalled = self.bike_key in (keys.LEFT, keys.RIGHT) and self.bike_key != self.last_bike_key
+            pedalled = self.bike_key in keys.LEFT + keys.RIGHT and self.bike_key != self.last_bike_key
             if pedalled:
                 self.started_pedalling = True
                 self.pedal()
@@ -503,30 +572,35 @@ class Eric(Character):
         self.x += self.direction
 
     def dismounted(self):
-        return self.sitting_on_saddle and self.bike_key == keys.DOWN
+        return self.sitting_on_saddle and self.bike_key in keys.DOWN
 
     def dismount(self):
         self.on_bike = False
 
     def stood_on_saddle(self):
-        if self.sitting_on_saddle and self.bike_key == keys.UP:
-            self.last_bike_key = keys.UP
+        if self.sitting_on_saddle and self.bike_key in keys.UP:
+            self.last_bike_key = self.bike_key
             self.sitting_on_saddle = False
         return not self.sitting_on_saddle
 
-    def stand_on_saddle(self, bike):
+    def stand_on_saddle(self):
         self.animatory_state = WALK0
-        self.x = bike.x
-        self.y = bike.y - 1
+        self.x -= self.direction
+        self.y -= 1
 
     def got_back_on_saddle(self):
-        if not self.sitting_on_saddle and self.bike_key == keys.DOWN:
-            self.last_bike_key = keys.DOWN
+        if not self.sitting_on_saddle and self.bike_key in keys.DOWN:
+            self.last_bike_key = self.bike_key
             self.sitting_on_saddle = True
         return self.sitting_on_saddle
 
+    def get_back_on_saddle(self):
+        self.animatory_state = RIDING_BIKE0
+        self.x += self.direction
+        self.y += 1
+
     def jumped(self):
-        return not self.sitting_on_saddle and self.bike_key in (keys.JUMP, keys.UP)
+        return not self.sitting_on_saddle and self.bike_key in keys.JUMP + keys.UP
 
     def check_cup(self):
         if items.FROG in self.inventory:
