@@ -17,11 +17,28 @@
 
 import pygame
 import random
+from barrier import Door, Wall, Window
+from floor import Floor
+from plant import PlantPot
+from lesson import AssemblyMessageGenerator
+from mutable import Shield, Safe, Cup, Bike
+from room import Room, Chair, Desk, Blackboard, NoGoZone
+from staircase import Staircase
+from scoreboard import Scoreboard
+from signals import Signals
+from location import Location
 import items
+import debug
+
+# Maximum time allowed to leave classroom or playground after bell rings
+GET_ALONG_TIME = 200
+
+# The command list Mr Wacker should use to expel Eric after he's exceeded the
+# lines limit
+TOO_MANY_LINES = 'ExpelEric:TooManyLines'
 
 class Skool:
-    def __init__(self, game, screen, beeper, cast, timetable, gallery):
-        self.game = game
+    def __init__(self, screen, beeper, cast, timetable, gallery, max_lines):
         self.screen = screen
         self.beeper = beeper
         self.scale = screen.scale
@@ -54,11 +71,12 @@ class Skool:
         self.draw_index = 0
         self.signals = Signals()
         self.messages = {}
-        self.expelled = False
         self.cups = {}
         self.plant_pots = {}
         self.bike = None
         self.clear_score = True
+        self.max_lines = max_lines
+        self.game_over = False
 
     def add_location(self, location_id, coords):
         self.locations[location_id] = coords
@@ -259,7 +277,7 @@ class Skool:
             self.end_game(False)
 
     def print_up_a_year_message(self):
-        message_lines = self.messages['NEXT_YEAR'].split('^')
+        message_lines = self.get_message('NEXT_YEAR')
         self.screen.print_lesson(*message_lines)
 
     def get_blackboards(self):
@@ -326,6 +344,12 @@ class Skool:
                 return staircase
 
     def next_staircase(self, home_floor, dest_floor):
+        if home_floor is None:
+            debug.log('Cannot determine next staircase when home floor is None')
+            return None
+        if dest_floor is None:
+            debug.log('Cannot determine next staircase when destination floor is None')
+            return None
         if home_floor is dest_floor:
             return None
         routes = self.routes[home_floor.floor_id]
@@ -356,6 +380,9 @@ class Skool:
     def in_no_go_zone(self, character):
         return any(zone.contains(character) for zone in self.no_go_zones)
 
+    def in_playground(self, character):
+        return 95 < character.x < 159
+
     def barrier(self, character, distance=0):
         for barrier in self.barriers.values():
             if barrier.impedes(character, distance):
@@ -403,6 +430,11 @@ class Skool:
         images, coords = self.barriers[barrier_id].move(shut)
         self.draw_mutable(images, coords)
 
+    def move_bike_away(self, barrier_id):
+        door = self.doors.get(barrier_id)
+        if door:
+            self.cast.move_bike_away(door)
+
     def auto_shut_doors(self):
         for barrier_id, door in self.doors.items():
             if door.auto_shut() and not self.cast.somebody_near_door(door):
@@ -507,7 +539,6 @@ class Skool:
         self.initialise_storeroom_combo()
 
     def reinitialise(self):
-        self.expelled = False
         self.screen.reinitialise()
         self.signals.clear()
         self.cast.reinitialise()
@@ -529,6 +560,7 @@ class Skool:
         for window in self.windows.values():
             self.move_door(window.barrier_id, door.initially_shut)
         self.chain_bike()
+        self.game_over = False
 
     def get_desks(self):
         desks = []
@@ -584,6 +616,9 @@ class Skool:
         """Sets the room Eric's supposed to be in at the moment."""
         self.home_room = self.get_room(self.timetable.get_room_id())
 
+    def unset_home_room(self):
+        self.home_room = None
+
     def get_home_room(self):
         """Returns the room Eric's supposed to be in at the moment."""
         return self.home_room
@@ -592,14 +627,22 @@ class Skool:
         """Return whether Eric is somewhere other than he should be."""
         if self.home_room:
             return not self.home_room.contains(eric)
-        if self.timetable.is_time(200):
+        if self.timetable.is_time(GET_ALONG_TIME):
             # Give Eric time to leave a classroom
             room = self.room(eric)
             if room:
                 destination = self.get_room(self.timetable.get_room_id())
                 return room != destination and room.get_along
+            elif self.in_playground(eric) and not self.is_playtime():
+                return True
         return False
 
+    def is_playtime(self):
+        return self.timetable.is_playtime()
+
+    def is_assembly(self):
+        return self.timetable.is_assembly()
+
     def get_lesson_desc(self):
         teacher_id = self.timetable.get_teacher_id()
         if self.timetable.hide_teacher():
@@ -637,15 +680,11 @@ class Skool:
         self.scoreboard.add_to_score(addend)
 
     def add_lines(self, addend):
-        if not self.expelled:
+        if not self.cast.is_eric_expelled():
             self.scoreboard.add_lines(addend)
-            if self.scoreboard.get_lines() > self.game.get_max_lines():
-                self.expelled = True
+            if self.scoreboard.get_lines() > self.max_lines:
                 self.stop_clock()
-                self.cast.expel_eric()
-
-    def is_eric_expelled(self):
-        return self.expelled
+                self.cast.expel_eric(TOO_MANY_LINES)
 
     def stop_clock(self):
         self.timetable.stop()
@@ -656,11 +695,12 @@ class Skool:
     def rewind_clock(self, ticks):
         self.timetable.rewind(ticks)
 
-    def get_too_many_lines_message(self):
-        return self.cast.expand_names(self.messages['TOO_MANY_LINES'])
+    def get_message(self, message_id):
+        message = self.expand_message(self.messages[message_id])
+        return message[0] if len(message) == 1 else message
 
-    def get_understand_message(self):
-        return self.messages['UNDERSTOOD']
+    def expand_message(self, message):
+        return self.cast.expand_names(message).split('^')
 
     def signal(self, signal):
         self.signals.signal(signal)
@@ -673,545 +713,4 @@ class Skool:
 
     def end_game(self, clear_score=True):
         self.clear_score = clear_score
-        self.game.end()
-
-class Location:
-    def __init__(self, coords):
-        self.x = coords[0]
-        self.y = coords[1]
-
-    def coords(self):
-        return self.x, self.y
-
-    def __str__(self):
-        return '(%s, %s)' % (self.x, self.y)
-
-class Staircase:
-    def __init__(self, bottom, top, force=False):
-        self.bottom = Location(bottom)
-        self.top = Location(top)
-        # Whether character *must* go up or down staircase when at bottom or
-        # top (instead of continuing left or right)
-        self.force = force
-        self.direction = 1 if self.bottom.x < self.top.x else -1
-
-    def contains(self, character, distance=0):
-        """
-        Returns whether the character is one of the following:
-            i) on a step of this staircase
-           ii) at the bottom of this staircase facing the top
-          iii) at the top of this staircase facing the bottom
-        """
-        for x, y in [(character.x + d * character.direction, character.y) for d in range(distance + 1)]:
-            if self.contains_location(x, y):
-                if x == self.bottom.x:
-                    return character.direction == self.direction
-                if x == self.top.x:
-                    return character.direction != self.direction
-                return True
-        return False
-
-    def contains_location(self, x, y):
-        """
-        Returns whether the given location (x, y) is at the bottom, at the top,
-        or on a step of this staircase.
-        """
-        if self.direction * self.bottom.x <= self.direction * x <= self.direction * self.top.x:
-            if (x, y) in (self.bottom.coords(), self.top.coords()):
-                return True
-            return self.top.y < y < self.bottom.y
-        return False
-
-    def supports(self, character):
-        """Returns whether the character is on a step of this staircase."""
-        if not self.contains_location(character.x, character.y):
-            return False
-        return (character.x, character.y) not in (self.bottom.coords(), self.top.coords())
-
-class Floor:
-    def __init__(self, floor_id, left_x, right_x, y):
-        self.floor_id = floor_id
-        self.left_x = left_x
-        self.right_x = right_x
-        self.y = y
-
-    def supports(self, character):
-        return self.y == character.y and self.left_x <= character.x <= self.right_x
-
-    def below(self, character):
-        return self.y >= character.y and self.left_x <= character.x <= self.right_x
-
-    def contains_location(self, x, y):
-        return self.y == y and self.left_x <= x <= self.right_x
-
-class Barrier:
-    def __init__(self, barrier_id, x, bottom_y, top_y):
-        self.barrier_id = barrier_id
-        self.x = x
-        self.bottom_y = bottom_y
-        self.top_y = top_y
-
-    def impedes(self, character, distance=0, force_shut=False):
-        if not (force_shut or self.is_shut()):
-            return False
-        if character.impeded(self.bottom_y, self.top_y):
-            if character.x < self.x and character.direction > 0:
-                return character.x + distance >= self.x - 2
-            elif character.x >= self.x and character.direction < 0:
-                return character.x - distance <= self.x
-        return False
-
-    def is_shut(self):
-        return True
-
-    def is_door(self):
-        return False
-
-class Wall(Barrier):
-    def separates(self, a, b):
-        """Returns whether this wall blocks the view from a to b."""
-        min_x = min(a.x, b.x)
-        max_x = max(a.x, b.x)
-        if min_x < self.x <= max_x:
-            return self.top_y <= max(a.y, b.y) and self.bottom_y >= min(a.y, b.y)
-
-class Door(Barrier):
-    def __init__(self, door_id, x, bottom_y, top_y, shut, auto_shuts):
-        Barrier.__init__(self, door_id, x, bottom_y, top_y)
-        self.shut = self.initially_shut = shut
-        self.auto_shuts = auto_shuts
-        self.open_images = None
-        self.shut_images = None
-        self.top_left = None
-        self.auto_shut_timer = 0
-        self.auto_shut_delay = 40
-
-    def is_shut(self):
-        return self.shut
-
-    def move(self, shut):
-        self.shut = shut
-        if not shut:
-            self.auto_shut_timer = self.auto_shut_delay
-        images = self.shut_images if shut else self.open_images
-        return (images, self.top_left)
-
-    def set_images(self, open_images, shut_images, top_left):
-        self.open_images = open_images
-        self.shut_images = shut_images
-        self.top_left = top_left
-
-    def auto_shut(self):
-        if not self.shut and self.auto_shuts:
-            self.auto_shut_timer -= 1
-        return self.auto_shut_timer < 0
-
-    def is_door(self):
-        return True
-
-class Window(Door):
-    def __init__(self, window_id, x, bottom_y, top_y, shut, opener_coords):
-        Door.__init__(self, window_id, x, bottom_y, top_y, shut, False)
-        self.opener_coords = opener_coords
-
-class Room:
-    def __init__(self, room_id, name, y, min_x, max_x, get_along):
-        self.room_id = room_id
-        self.name = name
-        self.y = y
-        self.min_x = min_x
-        self.max_x = max_x
-        self.get_along = get_along
-        self.chairs = []
-        self.desks = []
-        self.blackboard = None
-
-    def get_blackboard(self):
-        return self.blackboard
-
-    def get_blackboard_writer(self):
-        if self.blackboard:
-            return self.blackboard.get_writer()
-
-    def has_blackboard(self):
-        return self.blackboard is not None
-
-    def add_blackboard(self, x, y, font, scale):
-        self.blackboard = Blackboard(x, y, font, scale)
-
-    def blackboard_dirty(self):
-        if self.blackboard:
-            return self.blackboard.is_dirty()
-
-    def wipe_blackboard(self):
-        if self.blackboard:
-            self.blackboard.clear()
-
-    def get_blackboard_images(self):
-        if self.blackboard:
-            return self.blackboard.get_images()
-        return (0, 0, [])
-
-    def add_chair(self, x):
-        self.chairs.append(Chair(self, x))
-
-    def add_desk(self, x):
-        self.desks.append(Desk(self, x))
-
-    def beside_blackboard(self, character):
-        if self.blackboard:
-            return self.y == character.y and self.blackboard.beside(character)
-
-    def contains(self, character):
-        return self.y - 3 <= character.y <= self.y and self.min_x <= character.x <= self.max_x
-
-    def chair(self, character, direction=-1):
-        for chair in self.chairs:
-            if (direction is None or character.direction == direction) and (character.x, character.y) == (chair.x, chair.y):
-                return chair
-
-    def desk(self, character):
-        if character.is_sitting_on_chair():
-            for desk in self.desks:
-                if (character.x, character.y) == (desk.x, desk.y):
-                    return desk
-
-    def get_next_chair(self, character, move_along):
-        """Returns the chair nearest to the character, or the chair he should
-        sit in if he was dethroned (move_along=True) or has just started looking
-        for a chair (in which case he's facing right)."""
-        if character.direction > 0 or (move_along and self.chairs[0].x == character.x):
-            return self.chairs[-1]
-        min_distance = 100
-        next_chair = None
-        for chair in self.chairs:
-            distance = character.x - chair.x
-            if distance == 0 and move_along:
-                continue
-            if 0 <= distance < min_distance:
-                min_distance = distance
-                next_chair = chair
-        return next_chair
-
-class Chair:
-    def __init__(self, room, x):
-        self.x = x
-        self.y = room.y
-        self.room = room
-        self.occupant = None
-
-    def seat(self, character):
-        self.occupant = character
-
-    def vacate(self):
-        self.occupant = None
-
-class Desk:
-    def __init__(self, room, x):
-        self.x = x
-        self.y = room.y
-        self.room = room
-        self.empty()
-
-    def insert(self, item):
-        self.contents = item
-
-    def empty(self):
-        self.contents = None
-
-class Blackboard:
-    def __init__(self, x, y, font, scale):
-        self.x = x
-        self.y = y
-        self.font = font
-        self.width = 64 * scale
-        self.clear()
-
-    def set_writer(self, character):
-        self.writer = character
-
-    def get_writer(self):
-        return self.writer
-
-    def write(self, char):
-        if char == '^':
-            if len(self.lines) == 1:
-                self.lines.append('')
-            return
-        self.lines[-1] += char
-        self.images = []
-        for line in self.lines:
-            text_image = self.font.render(line, (255, 255, 255), (0, 0, 0))
-            if text_image.get_width() > self.width:
-                text_image = text_image.subsurface((0, 0), (self.width, text_image.get_height()))
-            self.images.append(text_image)
-
-    def wipe(self, wiped_bit, x):
-        for image in self.images:
-            image.blit(wiped_bit, (x, 0))
-
-    def is_dirty(self):
-        return len(self.images) > 0
-
-    def get_images(self):
-        return (self.x, self.y, self.images)
-
-    def clear(self):
-        self.lines = ['']
-        self.images = []
-        self.writer = None
-
-    def beside(self, character):
-        arm_x = character.x + character.direction + 1
-        return self.x + 1 <= arm_x <= self.x + 6
-
-    def shows(self, text, in_order=True):
-        if in_order:
-            return self.lines[0].lower().startswith(text.lower())
-        return set(text.lower()) == set(self.lines[0][:len(text)].lower())
-
-class NoGoZone:
-    def __init__(self, zone_id, min_x, max_x, bottom_y, top_y):
-        self.zone_id = zone_id
-        self.min_x = min_x
-        self.max_x = max_x
-        self.bottom_y = bottom_y
-        self.top_y = top_y
-
-    def contains(self, character):
-        if self.min_x <= character.x <= self.max_x:
-            return self.top_y <= character.y <= self.bottom_y
-        return False
-
-class Flashable:
-    def __init__(self, x, y, score, image=None, inverse_image=None):
-        self.x = x
-        self.y = y
-        self.score = score
-        self.image = image
-        self.inverse_image = inverse_image
-        self.flashing = False
-
-    def get_score(self):
-        return self.score // 10
-
-    def is_flashing(self):
-        return self.flashing
-
-    def flash(self):
-        self.flashing = True
-
-    def unflash(self):
-        self.flashing = False
-
-    def get_images(self, inverse):
-        if self.flashing:
-            image = self.inverse_image if inverse else self.image
-            return (self.x, self.y, image)
-        return (0, 0, None)
-
-class Shield(Flashable):
-    pass
-
-class Safe(Flashable):
-    pass
-
-class Cup:
-    def __init__(self, cup_id, coords):
-        self.cup_id = cup_id
-        self.x, self.y = coords
-        self.contents = None
-        self.frogs = []
-
-    def set_images(self, empty_images, water_images, sherry_images):
-        self.empty_images = empty_images
-        self.water_images = water_images
-        self.sherry_images = sherry_images
-
-    def is_empty(self):
-        return not (self.contents or self.frogs)
-
-    def fill(self, contents):
-        self.contents = contents
-        if contents == 'WATER':
-            images = self.water_images
-        elif contents == 'SHERRY':
-            images = self.sherry_images
-        else:
-            images = self.empty_images
-        return images, (self.x, self.y)
-
-    def insert_frog(self, frog):
-        self.frogs.append(frog)
-        frog.insert_into_cup(self)
-
-    def remove_frog(self, frog):
-        if frog in self.frogs:
-            self.frogs.remove(frog)
-
-class PlantPot:
-    def __init__(self, plant_id, plant, x, y):
-        self.plant_id = plant_id
-        self.plant = plant
-        self.x = x
-        self.y = y
-
-class Bike:
-    def __init__(self, x, y):
-        self.x = x
-        self.y = y
-
-    def set_images(self, unchained_images, chained_images):
-        self.unchained_images = unchained_images
-        self.chained_images = chained_images
-
-    def unchain(self):
-        return self.unchained_images, (self.x, self.y)
-
-    def chain(self):
-        return self.chained_images, (self.x, self.y)
-
-class Timetable:
-    def __init__(self, lesson_length, lesson_start_time):
-        self.index = -1
-        self.lesson_id = None
-        self.lessons = []
-        self.lesson_details = {}
-        self.counter = 0
-        self.ticking = True
-        self.lesson_length = lesson_length
-        self.lesson_start_time = lesson_start_time
-        self.special_playtimes = []
-
-    def reinitialise(self):
-        self.counter = 0
-        self.ticking = True
-        self.index = (self.index & 48) + 15
-
-    def add_lesson(self, lesson_id):
-        self.lessons.append(lesson_id)
-
-    def add_lesson_details(self, lesson_id, hide_teacher, teacher_id, room_id):
-        self.lesson_details[lesson_id] = (hide_teacher, teacher_id, room_id)
-
-    def add_special_playtime(self, lesson_id):
-        self.special_playtimes.append(lesson_id)
-
-    def next_lesson(self):
-        self.ticking = True
-        self.counter = self.lesson_length
-        self.index = (self.index + 1) % len(self.lessons)
-        self.lesson_id = self.lessons[self.index]
-        if self.lesson_id.startswith('Playtime') and self.special_playtimes and random.randrange(8) < 3:
-            self.lesson_id = random.choice(self.special_playtimes)
-
-    def get_lesson_id(self):
-        return self.lesson_id
-
-    def hide_teacher(self):
-        return self.lesson_details[self.get_lesson_id()][0]
-
-    def get_teacher_id(self):
-        return self.lesson_details[self.get_lesson_id()][1]
-
-    def get_room_id(self):
-        return self.lesson_details[self.get_lesson_id()][2]
-
-    def tick(self):
-        if self.ticking:
-            self.counter -= 1
-        return self.counter < 0
-
-    def is_time(self, time):
-        return self.counter + time < self.lesson_length
-
-    def is_time_remaining(self, time):
-        return self.counter < time
-
-    def is_time_to_start_lesson(self):
-        return self.is_time(self.lesson_start_time)
-
-    def is_teaching_eric(self, character):
-        return self.get_teacher_id() == character.character_id
-
-    def up_a_year(self):
-        self.counter = self.lesson_length // 2
-
-    def stop(self):
-        self.ticking = False
-
-    def resume(self, ticks):
-        self.counter = ticks
-        self.ticking = True
-
-    def rewind(self, ticks):
-        self.counter += ticks
-
-class AssemblyMessageGenerator:
-    def __init__(self):
-        self.text = None
-        self.verbs = []
-        self.nouns = []
-
-    def set_text(self, text):
-        self.text = text
-
-    def add_verb(self, verb):
-        self.verbs.append(verb)
-
-    def add_noun(self, noun):
-        self.nouns.append(noun)
-
-    def generate_message(self):
-        verb = random.choice(self.verbs)
-        noun = random.choice(self.nouns)
-        message = self.text.replace('$VERB', verb)
-        return message.replace('$NOUN', noun)
-
-class Scoreboard:
-    def __init__(self, screen):
-        self.screen = screen
-        self.hiscore = 0
-        self.reset()
-
-    def reset(self):
-        self.score = self.lines = 0
-
-    def add_to_score(self, addend):
-        self.score += addend
-        self.screen.print_score(self.score)
-
-    def add_lines(self, addend):
-        self.lines = max(self.lines + addend, 0)
-        self.screen.print_lines(self.lines)
-
-    def reinitialise(self):
-        if self.score > self.hiscore:
-            self.hiscore = self.score
-            self.screen.print_hi_score(self.hiscore)
-        self.reset()
-        self.screen.print_score(self.score)
-        self.screen.print_lines(self.lines)
-
-    def get_lines(self):
-        return self.lines
-
-class Signals:
-    ASSEMBLY_FINISHED = 'AssemblyFinished'
-    TIME_FOR_ASSEMBLY = 'TimeForAssembly'
-    SWOT_READY = 'SwotReady'
-
-    def __init__(self):
-        self.clear()
-
-    def clear(self):
-        self.signals = {}
-
-    def is_raised(self, signal):
-        return self.signals.get(signal, False)
-
-    def signal(self, signal):
-        self.signals[signal] = True
-
-    def unsignal(self, signal):
-        self.signals[signal] = False
+        self.game_over = True
